#include <iostream>
using namespace std;

#include <fstream>

#include <vector>
#include <cmath>
#include <numeric> 
#include <cstdlib>

#include "mpreal.h"
using namespace mpfr;

#include "Brutus.h"

int main(int argc, char* argv[]) {
  return 0;
}



    

